<?php

namespace Drupal\config_layers\EventSubscriber;

use Drupal\Core\Config\ConfigCrudEvent;
use Drupal\Core\Config\ConfigEvents;
use Drupal\Core\Database\Database;
use Drupal\config_layers\ConfigLayerManager;
use Drupal\config_layers\Entity\ConfigLayer;
use Drupal\config_layers\Config\FilteredStorage;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Config subscriber to allow automatic import of active configuration into layer.
 */
class ConfigSubscriber implements EventSubscriberInterface {

  /**
   * The Drupal Kernel.
   *
   * @var \Drupal\Core\ConfigLayerManager
   */
  protected $layerManager;

  /**
   * ConfigSubscriber constructor.
   *
   * @param \Drupal\config_layers\ConfigLayerManager $layerManager
   *   The layer manager.
   */
  public function __construct(ConfigLayerManager $layerManager) {
    $this->layerManager = $layerManager;
  }

  /**
   * Import new configuration into layers listening to events.
   *
   * @param \Drupal\Core\Config\ConfigCrudEvent $event
   *   The event to process.
   * @param string $eventName
   *   Name of the event.
   */
  public function importConfig(ConfigCrudEvent $event, $eventName) {
    $saved_config = $event->getConfig();
    $config_name = $saved_config->getName();
    $keys = [$config_name];

    $layer_ids = $this->layerManager->getLayers();
    $layers = ConfigLayer::loadMultiple($layer_ids);

    foreach ($layers as $layer) {
      if (in_array($eventName, $layer->getImportEvents())) {
        switch ($eventName) {
          case ConfigEvents::SAVE:
            // Optimization: import only the saved key.
            $source = new FilteredStorage($layer->getSourceStorage(), $keys);
            $layer->import($source, ConfigLayer::UPDATE_MODE_COPY);
            break;

          case ConfigEvents::DELETE:
            // Optimization: delete only the key.
            $layer->getDatabaseStorage()->delete($config_name);
            break;

          default:
            $keys = NULL;
            $layer->import();
        }
      }
    }

    $this->layerManager->mergeLayers($layers, $keys);
  }

  /**
   * Export configuration to layers listening to events.
   *
   * @param \Drupal\Core\Config\ConfigCrudEvent $event
   *   The event to process.
   * @param string $eventName
   *   Name of the event.
   */
  public function exportConfig(ConfigCrudEvent $event, $eventName) {
    $layer_ids = $this->layerManager->getLayers();
    $layers = ConfigLayer::loadMultiple($layer_ids);
    foreach ($layers as $layer) {
      if (in_array($eventName, $layer->getExportEvents())) {
        $layer->export();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    // Container has not been initialized at this stage,
    // i.e. we can not use \Drupal::service('config_layers.manager').
    // Workaround: query the db directly.
    $connection = Database::getConnection();
    if (isset($connection)) {
      $query = $connection->query("SELECT data FROM {config} WHERE name LIKE 'config_layers.layer.%'");
      $result = $query->fetchAll();

      $import_events = [];
      $export_events = [];

      if ($result) {
        foreach ($result as $row) {
          $data = unserialize($row->data);
          $import_events += $data['import_events'];
          $export_events += $data['export_events'];
        }
      }

      $events = [];
      foreach (array_unique($import_events) as $event) {
        $events[$event][] = ['importConfig', 10];
      }
      foreach (array_unique($export_events) as $event) {
        $events[$event][] = ['exportConfig', -10];
      }

      return $events;
    }
    return [];
  }

}
